/*-
 * Copyright (c) 2005 Daniel Braniss <danny@cs.huji.ac.il>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/*
 | $Id: fsm.c,v 1.20 2005/06/03 12:02:23 danny Exp danny $
 */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sysctl.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#if __FreeBSD_version < 500000
#include <sys/time.h>
#endif
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "iscsi.h"
#include "iscontrol.h"
#include "pdu.h"

static char *status_class1[] = {
     "Initiator error",
     "Authentication failure",
     "Authorization failure",
     "Not found",
     "Target removed",
     "Unsupported version",
     "Too many connections",
     "Missing parameter",
     "Can't include in session",
     "Session type not suported",
     "Session does not exist",
     "Invalid during login",
};
#define CLASS1_ERRS ((sizeof status_class1) / sizeof(char *))

static char *status_class3[] = {
     "Target error",
     "Service unavailable",
     "Out of resources"
};
#define CLASS3_ERRS ((sizeof status_class3) / sizeof(char *))

// XXX; should be static, but in the meantime ...  
int
handleLoginResp(isess_t *sess, pdu_t *pp)
{
     login_rsp_t *rp = (login_rsp_t *)pp;
     uint	st_class, status = ntohs(rp->status);

     debug_called(3);
     debug(4, "Tbit=%d csg=%d nsg=%d status=%x", rp->T, rp->CSG, rp->NSG, status);

     st_class  = status >> 8;
     if(status) {
	  int	st_detail = status & 0xff;

	  switch(st_class) {
	  case 1: // Redirect
	       switch(st_detail) {
		    // the ITN (iSCSI target Name) request a: 
	       case 1: // temporary address change
	       case 2: // permanent address change
		    status = 0;
	       }
	       break;

	  case 2: // Initiator Error
	       if(st_detail < CLASS1_ERRS)
		    printf("0x%04x: %s\n", status, status_class1[st_detail]);
	       break;

	  case 3:
	       if(st_detail < CLASS3_ERRS)
		    printf("0x%04x: %s\n", status, status_class3[st_detail]);
	       break;
	  }
     }
	  
     if(status == 0) {
	  processParams(sess, pp);
	  if(rp->T) {
	       sess->csg = rp->NSG;
	       if(sess->csg == FF_PHASE) {
		    // XXX: will need this when iplementing reconnect.
		    sess->tsih = rp->tsih;
		    debug(2, "TSIH=%x", sess->tsih);
	       }
	  }
     }

     return st_class;
}

static char *
getkeyval(char *key, pdu_t *pp)
{
    char	*ptr;
    int	klen, len, n;

    debug_called(3);

    len = pp->ds_len;
    ptr = pp->ds;
    klen = strlen(key);
    while(len > klen) {
	 if(strncmp(key, ptr, klen) == 0)
	      return ptr+klen;
	 n = strlen(ptr) + 1;
	 len -= n;
	 ptr += n;
    }
    return 0;
}

static int
handleChap(isess_t *sess, pdu_t *pp)
{
     pdu_t		spp;
     login_req_t	*lp;
     isc_opt_t		*op = sess->op;
     char		*ap, *ip, *cp, *digest; // MD5 is 128bits, SHA1 160bits

     debug_called(3);

     bzero(&spp, sizeof(pdu_t));
     lp = (login_req_t *)&spp.ipdu.bhs;
     lp->cmd = ISCSI_LOGIN_CMD | 0x40; // login request + Inmediate
     memcpy(lp->isid, sess->isid, 6);
     lp->tsih = sess->tsih;    // MUST be zero the first time!
     lp->CID = htons(1);
     lp->CSG = SN_PHASE;       // Security Negotiation
     lp->NSG = LON_PHASE;
     lp->T = 1;
    
     if(((ap = getkeyval("CHAP_A=", pp)) == NULL) ||
	((ip = getkeyval("CHAP_I=", pp)) == NULL) ||
	((cp = getkeyval("CHAP_C=", pp)) == NULL))
	  return -1;

     if((digest = chapDigest(ap, ip, cp, op->chapSecret)) == NULL)
	  return -1;

     addText(&spp, "CHAP_N=%s", op->chapIName? op->chapIName: op->initiatorName);
     addText(&spp, "CHAP_R=%s", digest);
     free(digest);

     return sendPDU(sess, &spp, handleLoginResp);
}

static int
authenticate(isess_t *sess)
{
     pdu_t		spp;
     login_req_t	*lp;
     isc_opt_t	*op = sess->op;

     bzero(&spp, sizeof(pdu_t));
     lp = (login_req_t *)&spp.ipdu.bhs;
     lp->cmd = ISCSI_LOGIN_CMD | 0x40; // login request + Inmediate
     memcpy(lp->isid, sess->isid, 6);
     lp->tsih = sess->tsih;	// MUST be zero the first time!
     lp->CID = htons(1);
     lp->CSG = SN_PHASE;	// Security Negotiation
     lp->NSG = SN_PHASE;
     lp->T = 0;

     switch((authm_t)lookup(AuthMethods, op->authMethod)) {
     case NONE:
	  return 0;

     case KRB5:
     case SPKM1:
     case SPKM2:
     case SRP:
	  return 2;

     case CHAP:
	  if(op->chapDigest == 0)
	       addText(&spp, "CHAP_A=5");
	  else
	  if(strcmp(op->chapDigest, "MD5") == 0)
	       addText(&spp, "CHAP_A=5,7");
	  else
	  if(strcmp(op->chapDigest, "SHA1") == 0)
	       addText(&spp, "CHAP_A=7,5");
	  else
	       addText(&spp, "CHAP_A=5");
	  return sendPDU(sess, &spp, handleChap);
     }
     return 1;
}

int
authenticateLogin(isess_t *sess)
{
     pdu_t		spp;
     login_req_t	*lp;
     int		status;

     debug_called(3);
     bzero(&spp, sizeof(pdu_t));
     lp = (login_req_t *)&spp.ipdu.bhs;
     lp->cmd = ISCSI_LOGIN_CMD | 0x40; // login request + Inmediate
     memcpy(lp->isid, sess->isid, 6);
     lp->tsih = sess->tsih;	// MUST be zero the first time!
     lp->CID = htons(1);
     lp->CSG = SN_PHASE;	// Security Negotiation
     lp->NSG = LON_PHASE;	// Login Operational Negotiation
     lp->T = 1;

     addText(&spp, "InitiatorName=%s", sess->op->initiatorName);
     addText(&spp, "SessionType=%s", sess->op->sessionType);
     if(strcmp(sess->op->sessionType, "Discovery") != 0)
	  addText(&spp, "TargetName=%s", sess->op->targetName);
     addText(&spp, "AuthMethod=%s", sess->op->authMethod);

     sess->csg = lp->CSG;
     status = sendPDU(sess, &spp, handleLoginResp);
     debug(3, "status=%d", status);
     if(status != 0)
	  return status;

     status = authenticate(sess);
     if(status == 0)
	  sess->flags |= SESS_LOGGEDIN;

     return status;
}
