/*-
 * Copyright (c) 2005 Daniel Braniss <danny@cs.huji.ac.il>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 | $Id: fsm.c,v 1.20 2005/06/03 12:02:23 danny Exp danny $
 */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sysctl.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#if __FreeBSD_version < 500000
#include <sys/time.h>
#endif
#include <sys/ioctl.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <syslog.h>
#include <stdarg.h>
#include <camlib.h>

#include "iscsi.h"
#include "iscontrol.h"
#include "pdu.h"

typedef enum {
     T1 = 1,
     T2, /*T3,*/ T4, T5, /*T6,*/ T7, T8, T9,
     T10, T11, T12, T13, T14, T15, T16, T18
} trans_t;

static trans_t
tcpConnect(isess_t *sess)
{
     isc_opt_t *op = sess->op;
     int	val;
     struct     addrinfo *res, hints;
     struct	sockaddr_in sn;
     struct	in_addr ipn;
     time_t	sec;

     debug_called(3);
     if(sess->flags & (SESS_RECONNECT|SESS_REDIRECT)) {
	  syslog(LOG_INFO, "%s", (sess->flags & SESS_RECONNECT)
		 ? "Reconnect": "Redirected");
	  
	  debug(3, "%s", (sess->flags & SESS_RECONNECT) ? "Reconnect": "Redirected");
	  close(sess->soc);
	  //sleep(2); // XXX:
	  sess->soc = -1;

	  sess->flags &= ~SESS_CONNECTED;
	  if(sess->flags & SESS_REDIRECT) {
	       if(sess->redirect_cnt++ > MAXREDIRECTS) {
		    syslog(LOG_WARNING, "too many redirects > %d", MAXREDIRECTS);
		    return 0;
	       }
	       sess->flags |= SESS_RECONNECT;
	  }
	  if((sess->flags & SESS_RECONNECT) == 0)
	       return 0;

	  // make sure we are not in a loop
	  // XXX: this code has to be tested
	  sec = time(0) - sess->reconnect_time;
	  if(sec > (5*60)) {
	       // if we've been connected for more that 5 minutes
	       // then just reconnect
	       sess->reconnect_time = sec;
	       sess->reconnect_cnt1 = 0;
	  }
	  else {
	       //
	       sess->reconnect_cnt1++;
	       if((sec / sess->reconnect_cnt1) < 2) {
		    // if less that 2 seconds from the last reconnect
		    // we are most probably looping
		    syslog(LOG_CRIT, "too many reconnects %d", sess->reconnect_cnt1);
		    return 0;
	       }
	  }
	  sess->reconnect_cnt++;
	  // sess->flags &= ~(SESS_RECONNECT|SESS_REDIRECT);
     }

     if((sess->soc = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  fprintf(stderr, "tcpConnect: socket: %m");
	  return 0;
     }

     memset(&hints, 0, sizeof(hints));
     hints.ai_family	= PF_INET;
     hints.ai_socktype	= SOCK_STREAM;

     debug(3, "targetAddress=%s port=%d", op->targetAddress, op->port);
     if(inet_aton(op->targetAddress, &ipn))
	  hints.ai_flags |= AI_NUMERICHOST;
     if((val = getaddrinfo(op->targetAddress, NULL, &hints, &res)) != 0) {
          fprintf(stderr, "getaddrinfo(%s): %s", op->targetAddress, gai_strerror(val));
          return 0;
     }
     memcpy(&sn, res->ai_addr, sizeof(struct sockaddr_in));
     sn.sin_port = htons(op->port);
     freeaddrinfo(res);

     // from Patrick.Guelat@imp.ch:
     // iscontrol can be called without waiting for the socket entry to time out
     val = 1;
     if(setsockopt(sess->soc, SOL_SOCKET, SO_REUSEADDR, &val, (socklen_t)sizeof(val)) < 0) {
	  fprintf(stderr, "Cannot set socket SO_REUSEADDR %d: %s\n\n",
		  errno, strerror(errno));
     }

     sess->flags &= ~SESS_CONNECTED;

     if(connect(sess->soc, (struct sockaddr *)&sn, sizeof(struct sockaddr_in)) != -1) {
	  struct	timeval timeout;

	  sess->flags |= SESS_CONNECTED;
#if 1
	  val = 1;
	  if(setsockopt(sess->soc, IPPROTO_TCP, TCP_NODELAY, &val, sizeof(val)) < 0)
	       fprintf(stderr, "Cannot set socket NO delay option err=%d %s\n",
		       errno, strerror(errno));
	  
	  val = 96 * 1024;
	  if((setsockopt(sess->soc, SOL_SOCKET, SO_SNDBUF, &val, sizeof(val)) < 0)
	     || (setsockopt(sess->soc, SOL_SOCKET, SO_RCVBUF, &val, sizeof(val)) < 0))
	       fprintf(stderr, "Cannot set socket sndbuf & rcvbuf to %d err=%d %s\n",
		       val, errno, strerror(errno));
	  
	  timeout.tv_sec = 10;
	  timeout.tv_usec = 0;
	  if((setsockopt(sess->soc, SOL_SOCKET, SO_SNDTIMEO, &timeout, sizeof(timeout)) < 0)
	     || (setsockopt(sess->soc, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout)) < 0)) {
	       fprintf(stderr, "Cannot set socket timeout to %ld err=%d %s\n",
		       timeout.tv_sec, errno, strerror(errno));
	  }
#endif
     } 
     else {
	  perror("connect");
     }
     return T1;
}

static trans_t
startSession(isess_t *sess)
{
     int	n, fd, nfd;
     char	*dev;

     debug_called(3);

     if((sess->flags & SESS_CONNECTED) == 0) {
	  return T2;
     }
     if(sess->fd == -1) {
	  fd = open(iscsidev, O_RDWR);
	  if(fd < 0) {
	       perror(iscsidev);
	       return 0;
	  }
	  n = 7;
	  if(sysctlbyname("hw.iscsi.isid", (void *)sess->isid, (size_t *)&n, 0, 0) != 0)
	       perror("sysctlbyname");
	  
	  if(ioctl(fd, ISCSISETSES, &n)) {
	       perror("ISCSISETSES");
	       return 0;
	  }
	  asprintf(&dev, "%s%d", iscsidev, n);
	  nfd = open(dev, O_RDWR);
	  if(nfd < 0) {
	       perror(dev);
	       free(dev);
	       return 0;
	  }
	  free(dev);
	  close(fd);
	  sess->fd = nfd;
     }

     if(ioctl(sess->fd, ISCSISETSOC, &sess->soc)) {
	  perror("ISCSISETSOC");
	  return 0;
     }

     if((sess->flags & SESS_FULLFEATURE) == 0) {
	  if(setOptions(sess) != 0)
	       return -1;
     }

     return T4;
}

static int
handledDiscoveryResp(isess_t *sess, pdu_t *pp)
{
     char	*ptr;
     int	len, n;

     debug_called(3);

     len = pp->ds_len;
     ptr = pp->ds;
     while(len > 0) {
	  printf("%s\n", ptr);
	  n = strlen(ptr) + 1;
	  len -= n;
	  ptr += n;
     }
     return 0;
}

static int
doDiscovery(isess_t *sess)
{
     pdu_t		spp;
     text_req_t	*tp = (text_req_t *)&spp.ipdu.bhs;

     debug_called(3);

     bzero(&spp, sizeof(pdu_t));
     tp->cmd = ISCSI_TEXT_CMD | 0x40;
     tp->F = 1;
     tp->ttt = 0xffffffff;
     addText(&spp, "SendTargets=All");
     return sendPDU(sess, &spp, handledDiscoveryResp);
}

static trans_t
doLogin(isess_t *sess)
{
     int status;

     debug_called(3);

     status = 0;
     sess->csg = SN_PHASE;
     do {
	  debug(3, "csg=%d", sess->csg);

	  switch(sess->csg) {
	  case SN_PHASE:	// Security Negotiation
	       status = authenticateLogin(sess);
	       if(status != 0)
		    goto bad;
	       break;
	       
	  case LON_PHASE:	// Login Operational Negotiation
	       status = negotiateOPV(sess);
	       if(status != 0)
		    goto bad;
	       break;
	       
	  case FF_PHASE:	// FuLL-Feature
	       if(!status && strcmp(sess->op->sessionType, "Discovery") == 0)
		    goto discovery;
	       
	       // we are done
	       return T5;
	  }
     } while(1);

 discovery:
     status = doDiscovery(sess);
 bad:
     debug(3, "status=%d", status);
     switch(status) {
     case 0:
	  return T5;

     case 1:	// redirect - temporary/permanent
	  sess->flags |= SESS_REDIRECT;
	  return T7;
     case 2: // initiator terminal error
     case 3: // target terminal error -- could retry ...
     default:
	  return 0;
     }
}

isess_t *currsess;

static void
trap(int sig)
{
     syslog(LOG_NOTICE, "trapped signal %d", sig);

     switch(sig) {
     case SIGHUP:
	  currsess->flags |= SESS_DISCONNECT;
	  break;

     case SIGUSR1:
	  currsess->flags |= SESS_RECONNECT;
	  break;

     default:
	  exit(0); // XXX: for shutdown ??
	  break;
     }
}

static void
doCAM(isess_t *sess)
{
     char	pathstr[1024];
     union ccb	*ccb;
     int	i;

     if(ioctl(sess->fd, ISCSIGETCAM, &sess->cam) != 0) {
	  syslog(LOG_WARNING, "ISCSIGETCAM failed: %d", errno);
	  return;
     }
     debug(2, "nluns=%d", sess->cam.target_nluns);
     /*
      | for now will do this for each lun ...
      */
     for(i = 0; i < sess->cam.target_nluns; i++) {
	  debug(2, "CAM path_id=%d target_id=%d target_lun=%d",
		sess->cam.path_id, sess->cam.target_id, sess->cam.target_lun[i]);

	  sess->camdev = cam_open_btl(sess->cam.path_id, sess->cam.target_id,
				      sess->cam.target_lun[i], O_RDWR, NULL);
	  if(sess->camdev == NULL) {
	       syslog(LOG_WARNING, "%s", cam_errbuf);
	       debug(3, "%s", cam_errbuf);
	       continue;
	  }

	  cam_path_string(sess->camdev, pathstr, sizeof(pathstr));
	  debug(2, "pathstr=%s", pathstr);

	  ccb = cam_getccb(sess->camdev);
	  bzero(&(&ccb->ccb_h)[1], sizeof(struct ccb_relsim) - sizeof(struct ccb_hdr));
	  ccb->ccb_h.func_code = XPT_REL_SIMQ;
	  ccb->crs.release_flags = RELSIM_ADJUST_OPENINGS;
	  ccb->crs.openings = sess->op->tags;

	  if(cam_send_ccb(sess->camdev, ccb) < 0)
	       syslog(LOG_WARNING, "%s", cam_errbuf);
	  else
	  if((ccb->ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_CMP) {
	       syslog(LOG_WARNING, "XPT_REL_SIMQ CCB failed");
	       // cam_error_print(sess->camdev, ccb, CAM_ESF_ALL, CAM_EPF_ALL, stderr);
	  }
	  else
	       syslog(LOG_INFO, "%s tagged openings now %d\n", pathstr, ccb->crs.openings);

	  cam_freeccb(ccb);
	  cam_close_device(sess->camdev);
     }
}

static trans_t
supervise(isess_t *sess)
{
     int	sig;

     debug_called(3);

     if(strcmp(sess->op->sessionType, "Discovery") == 0) {
	  sess->flags |= SESS_DISCONNECT;
	  return T9;
     }

     if(vflag)
	  printf("ready to go scsi\n");
     if((sess->flags & SESS_FULLFEATURE) == 0) {
	  if(daemon(0, 1) != 0) {
	       perror("daemon");
	       exit(1);
	  }

	  openlog("iscontrol", LOG_CONS|LOG_PERROR|LOG_PID|LOG_NDELAY, LOG_KERN);
	  syslog(LOG_INFO, "running");

	  currsess = sess;
	  
	  signal(SIGHUP, trap);
	  if(ioctl(sess->fd, ISCSISTART)) {
	       perror("ISCSISTART");
	       return -1;
	  }
	  sig = SIGUSR1;

	  signal(sig, trap);
	  if(ioctl(sess->fd, ISCSISIGNAL, &sig)) {
	       perror("ISCSISIGNAL");
	  }
	  sess->flags |= SESS_FULLFEATURE;
#if 0
	  signal(SIGTERM, trap);
	  signal(SIGINT, trap);
#endif
	  doCAM(sess);

     }
     sess->flags &= ~(SESS_REDIRECT | SESS_RECONNECT);
     printf("starting main loop\n");
     while((sess->flags & (SESS_REDIRECT|SESS_RECONNECT|SESS_DISCONNECT)) == 0) {
	  // do something?
	  // like sending a nop_out?
	  sleep(60);
     }

     syslog(LOG_INFO, "sess flags=%x", sess->flags);

     sig = 0;
     if(ioctl(sess->fd, ISCSISIGNAL, &sig)) {
	  perror("ISCSISIGNAL");
     }

     if(sess->flags & SESS_DISCONNECT) {
	  sess->flags &= ~SESS_FULLFEATURE;
	  return T9; // we are done
     }
     return T8;
}

static int
handleLogoutResp(isess_t *sess, pdu_t *pp)
{
     if(sess->flags & SESS_DISCONNECT)
	  return 0;
     return T13;
}

static trans_t
startLogout(isess_t *sess)
{
     pdu_t	spp;
     logout_req_t *p = (logout_req_t *)&spp.ipdu.bhs;

     bzero(&spp, sizeof(pdu_t));
     p->cmd = ISCSI_LOGOUT_CMD| 0x40;
     p->reason = 0;
     p->CID = htons(1);

     return sendPDU(sess, &spp, handleLogoutResp);
}

static trans_t
inLogout(isess_t *sess)
{
     return 0;
}

typedef enum {
     S1, S2, /*S3,*/ S4, S5, S6, S7, S8
} state_t;

int
fsm(isc_opt_t *op)
{
     state_t state = S1;
     isess_t	*sess;

     if((sess = calloc(1, sizeof(isess_t))) == NULL) {
	  // boy, is this a bad start ...
	  fprintf(stderr, "no memory!\n");
	  return -1;
     }

     sess->op = op;
     sess->fd = -1;
     sess->soc = -1;

     do switch(state) {

     case S1:
	  switch(tcpConnect(sess)) {
	  case T1: state = S2; break;
	  default: state = S8; break;
	  }
	  break;

     case S2:
	  switch(startSession(sess)) {
	  case T2: state = S1; break;
	  case T4: state = S4; break;
	  default: state = S8; break;
	  }
	  break;

     case S4:
	  switch(doLogin(sess)) {
	  case T7:  state = S1; break;
	  case T5:  state = S5; break;
	  default: state = S8; break;
	  }
	  break;

     case S5:
	  switch(supervise(sess)) {
	  case T8:  state = S1; break;
	  case T9:  state = S6; break;
	  case T11: state = S7; break;
	  case T15: state = S8; break;
	  default: state = S8; break;
	  }
	  break;

     case S6:
	  switch(startLogout(sess)) {
	  case T13: state = S1; break;
	  case T14: state = S6; break;
	  case T16: state = S8; break;
	  default: state = S8; break;
	  }
	  break;
	  
     case S7: 
	  switch(inLogout(sess)) {
	  case T18: state = S1; break;
	  case T10: state = S6; break;
	  case T12: state = S7; break;
	  case T16: state = S8; break;
	  default: state = S8; break;
	  }
	  break;

     case S8:
	  // maybe do some clean up?
	  syslog(LOG_INFO, "terminated");
	  return 0;

     } while(1);
}
