/*-
 * Copyright (c) 2005 Daniel Braniss <danny@cs.huji.ac.il>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/*
 | iSCSI
 | $Id: isc_soc.c,v 1.16 2005/09/09 09:44:38 danny Exp danny $
 */
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/conf.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/ctype.h>
#include <sys/errno.h>
#include <sys/sysctl.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/socketvar.h>
#include <sys/socket.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/ioccom.h>
#include <sys/queue.h>
#include <sys/kthread.h>
#include <sys/syslog.h>
#include <sys/mbuf.h>
#include <sys/user.h>
#include <unistd.h>

#include <dev/iscsi/iscsi.h>
#include <dev/iscsi/iscsivar.h>
/*
 | wait till a PDU header is received
 | from the socket.
 */
static int
so_getbhs(isc_session_t *sp)
{
     bhs_t *bhs		= &sp->bhs;
     struct uio		*uio = &sp->uio;
     struct iovec	*iov = &sp->iov;
     int		error, flags;

     debug_called(8);

     iov->iov_base	= bhs;
     iov->iov_len	= sizeof(bhs_t);

     uio->uio_iov	= iov;
     uio->uio_iovcnt	= 1;
     uio->uio_rw	= UIO_READ;
     uio->uio_segflg	= UIO_SYSSPACE;
     uio->uio_td	= curthread; // why ...
     uio->uio_resid	= sizeof(bhs_t);

     flags = MSG_WAITALL;
     error = soreceive(sp->soc, NULL, uio, 0, 0, &flags);
     if(!error && uio->uio_resid > 0) {
#if __FreeBSD_version <  600000
	  debug(5, "so_error=%d uio->uio_resid=%d iov.iov_len=%d",
#else
	  debug(5, "so_error=%d uio->uio_resid=%d iov.iov_len=%zd",
#endif
		sp->soc->so_error, uio->uio_resid, iov->iov_len);
	  error = EPIPE;
     }
     return error;
}

/*
 | so_recv gets called when there is at least
 | an iSCSI header in the queue
 */
static int
so_recv(isc_session_t *sp, pduq_t *pq)
{
     struct socket	*so = sp->soc;
     sn_t		*sn = &sp->sn;
     struct uio		*uio = &pq->uio;
     pdu_t		*pp;
     int		error;
     size_t		n, len;
     bhs_t		*bhs;
     u_int		max, exp;

     debug_called(8);
     /*
      | now calculate how much data should be in the buffer
      | NOTE: digest is not verified/calculated
      */
     pp = &pq->pdu;
     bhs = &pp->ipdu.bhs;

     len = 0;
     if(bhs->AHSLength) {
	  pp->ahs_len = bhs->AHSLength * 4;
	  if(sp->hdrDigest)
	       pp->ahs_len += 4;
	  len += pp->ahs_len;
     }
     if(bhs->DSLength) {
	  n = bhs->DSLength;
#ifdef LITTLE_ENDIAN
	  pp->ds_len = ((n & 0x00ff0000) >> 16)
	       | (n & 0x0000ff00)
	       | ((n & 0x000000ff) << 16);
#else
	  pp->ds_len = n;
#endif
	  len += pp->ds_len;
	  while(len & 03)
	       len++;
	  if(sp->dataDigest)
	       len += 4;
     }
     if(len > sp->opt.maxRecvDataSegmentLength) {
	  // deep trouble here, probably all we can do is
	  // force a disconnect, XXX: check RFC ...
#if 0
	  log(LOG_ERR,
	      "so_recv: impossible PDU length(%ld) from iSCSI %s/%s\n",
	      len, sp->opt.targetAddress, sp->opt.targetName);
#endif
	  error = E2BIG;
	  goto out;
     }
     if(len) {
	  int	flags;

	  uio->uio_resid = len;
	  uio->uio_td = curthread; // why ...
	  flags = MSG_WAITALL;

	  error = soreceive(so, NULL, uio, &pq->mp, NULL, &flags);
	  //if(error == EAGAIN)
	  // XXX: this needs work! it hangs iscontrol
	  if(error || uio->uio_resid)
	       goto out;
     }

     sdebug(5, "opcode=0x%x ahs_len=0x%x ds_len=0x%x",
	    bhs->opcode, pp->ahs_len, pp->ds_len);


     max = ntohl(bhs->MaxCmdSN);
     exp = ntohl(bhs->ExpStSN);


     if(max < exp - 1 &&
	max > exp - _MAXINCR) {
	  sdebug(2,  "bad cmd window size");
	  error = EIO; // XXX: for now;
	  goto out; // error
     }

     if(SNA_GT(max, sn->maxCmd))
	  sn->maxCmd = max;

     if(SNA_GT(exp, sn->expCmd))
	  sn->expCmd = exp;
     sp->cws = sn->maxCmd - sn->expCmd + 1;

     return 0;

 out:
     // XXX: need some work here
     xdebug("have a problem, error=%d", error);
     pdu_free(sp->isc, pq);
     if(!error && uio->uio_resid > 0)
	  error = EPIPE;
     return error;
}

int
isc_sendPDU(isc_session_t *sp, pduq_t *pq)
{
     pdu_t		*pp = &pq->pdu;
     struct uio		*uio = &pq->uio;
     struct iovec	*iv;
     int		len, error;

     debug_called(8);

     bzero(uio, sizeof(struct uio));

     uio->uio_rw	= UIO_WRITE;
     uio->uio_segflg	= UIO_SYSSPACE;
     uio->uio_td	= sp->td;
     uio->uio_iov 	= iv = pq->iov;

     iv->iov_base	= &pp->ipdu;
     iv->iov_len	= sizeof(union ipdu_u);
     uio->uio_resid	= pq->len;
     iv++;

     if(pp->ahs_len) {
	  iv->iov_base	= pp->ahs;
	  iv->iov_len	= pp->ahs_len;
	  iv++;
     }
     if(sp->hdrDigest) {
	  pq->pdu.hdr_dig = sp->hdrDigest(&pp->ipdu, uio->uio_resid);
	  iv->iov_base	= &pp->hdr_dig;
	  iv->iov_len	= sizeof(int);
	  iv++;
     }
     if(pq->pdu.ds) {
	  iv->iov_base	= pp->ds;
	  iv->iov_len	= pp->ds_len;
	  while(iv->iov_len & 03) // the specs say it must be int alligned
	       iv->iov_len++;
	  iv++;
     }
     if(sp->dataDigest) {
	  pp->ds_dig = sp->dataDigest(pp->ds, pp->ds_len);
	  iv->iov_base	= &pp->ds_dig;
	  iv->iov_len	= sizeof(int);
	  iv++;
     }
     uio->uio_iovcnt	= iv - pq->iov;
     sdebug(5, "opcode=%x iovcnt=%d uio_resid=%d itt=%x",
	    pp->ipdu.bhs.opcode, uio->uio_iovcnt, uio->uio_resid,
	    ntohl(pp->ipdu.bhs.itt));
     sp->stats.nsent++;
     do {
	  len = uio->uio_resid;
	  error = sosend(sp->soc, NULL, uio, 0, 0, 0, sp->td);
	  if(uio->uio_resid == 0 || error || len == uio->uio_resid) {
	       if(uio->uio_resid != 0) {
		    sdebug(3, "uio->uio_resid=%d uio->uio_iovcnt=%d error=%d len=%d",
			   uio->uio_resid, uio->uio_iovcnt, error, len);
	       }
	       break;
	  }
	  sdebug(1, "uio->uio_resid=%d uio->uio_iovcnt=%d",
		uio->uio_resid, uio->uio_iovcnt);
	  iv = uio->uio_iov;
	  len -= uio->uio_resid;
	  while(uio->uio_iovcnt > 0) {
	       if(iv->iov_len > len) {
		    caddr_t	bp = (caddr_t)iv->iov_base;

		    iv->iov_len -= len;
		    iv->iov_base = (void *)&bp[len];
		    break;
	       }
	       len -= iv->iov_len;
	       uio->uio_iovcnt--;
	       uio->uio_iov++;
	       iv++;
	  }
     } while(uio->uio_resid);

     return error;
}
/*
 | one per active session.
 | wait for something to arrive.
 | and if the pdu is without errors, process it.
 */
static void
isc_soc(void *vp)
{
     isc_session_t	*sp = (isc_session_t *)vp;
     struct socket	*so = sp->soc;
     pduq_t		*pq;
     int		error;

     debug_called(8);
     sdebug(3, "started");

     sp->flags &= ~(ISC_IWAITING|ISC_CON_STOPPED);
     sp->flags |= ISC_CON_RUN;

     wakeup(sp); // tell isc_start_receiver that all is ok.

     if(sp->cam_path)
	  ic_release(sp);

     error = 0;
     while((sp->flags & ISC_CON_RUN) && (so->so_state & SS_ISCONNECTED)) {
	  /*
	   | first read in the iSCSI header
	   */
	  error = so_getbhs(sp);
	  if(error) {
	       if(error == EPIPE)
		    break;
	  }
	  else {
	       pq = pdu_alloc(sp->isc, 1);  // OK to WAIT
	       pq->pdu.ipdu.bhs = sp->bhs;
	       if((error = so_recv(sp, pq)) != 0) {
		    error += 0x800; // XXX: just to see the error.
		    // terminal error
		    // XXX: close connection and exit
		    break;
	       }
	       ism_recv(sp, pq);

	       mtx_lock(&sp->io_mtx);
	       if(sp->flags & ISC_IWAITING) {
		    sp->flags &= ~ISC_IWAITING;
		    wakeup(&sp->flags);
	       }
	       mtx_unlock(&sp->io_mtx);
	       sp->stats.nrecv++;
	  }
     }
     sdebug(3, "terminated flags=%x so_count=%d so_state=%x error=%d",
	    sp->flags, so->so_count, so->so_state, error);

     if((sp->proc != NULL) && sp->signal) {
	  PROC_LOCK(sp->proc);
	  psignal(sp->proc, sp->signal);
	  PROC_UNLOCK(sp->proc);
	  sp->flags |= ISC_SIGNALED;
	  sdebug(1, "signaled");
     }
     else {
	  // we have to do something ourselfs
	  // like closing this session ...
     }

     if(sp->cam_path)
	  ic_freeze(sp);
     /*
      | we've been terminated
      */
     // do we need this mutex ...?
     mtx_lock(&sp->io_mtx);
     sp->flags |= ISC_CON_STOPPED;
     sp->flags &= ~ISC_CON_RUN;

     wakeup(sp);

     mtx_unlock(&sp->io_mtx);

     kthread_exit(0);
}

void
isc_stop_receiver(isc_session_t *sp)
{
     debug_called(8);

     mtx_lock(&sp->io_mtx);
     if(sp->flags & ISC_CON_RUN) {
	  int n = 5;

	  sp->flags &= ~ISC_CON_RUN;
	  while(n-- && (sp->flags & ISC_CON_STOPPED) == 0) {
	       sdebug(3, "waiting n=%d... flags=%x", n, sp->flags);
	       msleep(sp, &sp->io_mtx, PRIBIO, "isc_stpc", 5*hz);
	  }
	  sp->flags |= ISC_CON_STOPPED;
     }
     mtx_unlock(&sp->io_mtx);
}

void
isc_start_receiver(isc_session_t *sp)
{
     debug_called(8);

     sp->flags &= ~ISC_CON_STOPPED;

     kthread_create(isc_soc, sp, &sp->proc, 0, 0, "iscsi%d", sp->sid);

     mtx_lock(&sp->io_mtx);
     while((sp->flags & ISC_CON_RUN) == 0) {
	  sdebug(3, "waiting flags=%x", sp->flags);
	  msleep(sp, &sp->io_mtx, PRIBIO, "isc_stpc", 1*hz);
     }
     mtx_unlock(&sp->io_mtx);
}
